from fastapi import APIRouter, Request

from ..openai import Openai
from .schemas import OpenAIV1ChatCompletion

openai = Openai()
router = APIRouter(prefix=openai.ROUTE_PREFIX, tags=["v1"])


@router.post("/v1/chat/completions")
async def chat_completions(params: OpenAIV1ChatCompletion, request: Request):
    """该接口只是为了document, 将此路由接口放在openai.reverse_proxy接口后面, 实际不会执行该接口。"""
    return await openai.v1_chat_completions(params, request)



import httpx
from sparrow.api import create_app

from .openai import Openai
from .routers.openai_v1 import router as router_v1

app = create_app(title="openai_forward", version="1.0")
openai = Openai()

app.add_route(
    openai.ROUTE_PREFIX + '/{api_path:path}',
    openai.reverse_proxy,
    methods=['GET', 'POST', 'PUT', 'DELETE', 'OPTIONS', 'HEAD', 'PATCH', 'TRACE'],
)

app.include_router(router_v1)



from __future__ import annotations

from typing import Dict, List, Optional, Union

from pydantic import BaseModel, Field


class OpenAIV1ChatCompletion(BaseModel):
    """Creates a completion for the chat message"""

    model: str = Field(
        ..., description="The model to use for the completion", example="gpt-3.5-turbo"
    )
    messages: List[Dict[str, str]] = Field(
        ...,
        description="The message to complete",
        example=[{"role": "user", "content": "hi"}],
    )
    temperature: float = Field(default=1, description="0会导致更确定的结果，1会导致更随机的结果")
    top_p: float = Field(default=1, description="0会导致更确定的结果，1会导致更随机的结果")
    n: int = Field(
        default=1,
        description="How many chat completion choices to generate for each input message.",
    )
    stream: bool = Field(default=False)
    stop: Union[List[str], str, None] = Field(
        default=None,
        description="Up to 4 sequences where the API will stop generating further tokens.",
    )
    max_tokens: Union[int, None] = Field(
        default=None,
        description="The maximum number of tokens to generate in the chat completion. The total length of input tokens and generated tokens is limited by the model's context length.",
    )
    presence_penalty: float = Field(
        default=0, description="Number between -2.0 and 2.0. "
    )
    frequency_penalty: float = Field(
        default=0, description="Number between -2.0 and 2.0. "
    )
    logit_bias: Optional[Dict[str, float]] = Field(
        default=None,
        description="取值[-100, 100]， 取值越大，生成的结果越偏向于该token。 取-100表示完全不考虑该token。",
    )
    user: Optional[str] = Field(
        default=None,
        description="A unique identifier representing your end-user, which can help OpenAI to monitor and detect abuse. ",
    )

    class Config:
        schema_extra = {
            "example": {
                "model": "gpt-3.5-turbo",
                "messages": [{"role": "user", "content": "hi"}],
                "stream": False,
                "temperature": 1,
                "top_p": 1,
                "logit_bias": None,
            }
        }



import os

import fire
import uvicorn


class Cli:
    @staticmethod
    def run(
        port=8000,
        workers=1,
        api_key=None,
        forward_key=None,
        base_url=None,
        log_chat=None,
        route_prefix=None,
        ip_whitelist=None,
        ip_blacklist=None,
    ):
        """Run forwarding serve.

        Parameters
        ----------

        port: int, default 8000
        workers: int, 1
        api_key: str, None
        forward_key: str, None
        base_url: str, None
        log_chat: str, None
        route_prefix: str, None
        ip_whitelist: str, None
        ip_blacklist: str, None
        """
        if base_url:
            os.environ['OPENAI_BASE_URL'] = base_url
        if api_key:
            os.environ['OPENAI_API_KEY'] = api_key
        if forward_key:
            os.environ['FORWARD_KEY'] = forward_key
        if log_chat:
            os.environ['LOG_CHAT'] = log_chat
        if route_prefix:
            os.environ['ROUTE_PREFIX'] = route_prefix
        if ip_whitelist:
            os.environ['IP_WHITELIST'] = ip_whitelist
        if ip_blacklist:
            os.environ['IP_BLACKLIST'] = ip_blacklist

        ssl_keyfile = os.environ.get("ssl_keyfile", None) or None
        ssl_certfile = os.environ.get("ssl_certfile", None) or None
        uvicorn.run(
            app="openai_forward.app:app",
            host="0.0.0.0",
            port=port,
            workers=workers,
            app_dir='..',
            ssl_keyfile=ssl_keyfile,
            ssl_certfile=ssl_certfile,
        )


def main():
    fire.Fire(Cli)


if __name__ == "__main__":
    main()



import os
from pathlib import Path
from typing import Dict, List

import orjson
from fastapi import Request
from httpx._decoders import LineDecoder
from loguru import logger
from orjson import JSONDecodeError
from sparrow import relp

decoder = LineDecoder()


def _parse_iter_line_content(line: str):
    line = line[6:]
    try:
        line_dict = orjson.loads(line)
        return line_dict['choices'][0]['delta']['content']
    except JSONDecodeError:
        return ""
    except KeyError:
        return ""


def parse_chat_completions(bytes_: bytes):
    txt_lines = decoder.decode(bytes_.decode('utf-8'))
    line0 = txt_lines[0]
    target_info = dict()
    if line0.startswith("data:"):
        line0 = orjson.loads(line0[6:])
        msg = line0['choices'][0]['delta']
    else:
        line0 = orjson.loads(line0)
        msg = line0['choices'][0]['message']

    target_info['created'] = line0['created']
    target_info['id'] = line0['id']
    target_info['model'] = line0['model']
    target_info['role'] = msg['role']
    target_info['content'] = msg.get("content", "")
    # loop for stream
    for line in txt_lines[1:]:
        if line in ("", "\n", "\n\n"):
            continue
        elif line.startswith("data: "):
            target_info['content'] += _parse_iter_line_content(line)
        else:
            logger.warning(f"line not startswith data: {line}")
    return target_info


class ChatSaver:
    def __init__(self, max_chat_size=2000, save_interval=2, _dir='./Log'):
        self._chat_list = []
        self._file_idx = 0
        self._save_interval = save_interval
        self._max_chat_file_size = max_chat_size
        self._cur_chat_file_size = 0
        self._log_dir = _dir
        self._init_chat_file()

    @property
    def chat_file(self):
        return os.path.join(self._log_dir, f"chat_{self._file_idx}.txt")

    def _init_chat_file(self):
        _log_dir = Path(self._log_dir)
        if not _log_dir.exists():
            _log_dir.mkdir(parents=True)
        while Path(self.chat_file).exists():
            self._file_idx += 1

    @staticmethod
    async def parse_payload_to_content(request: Request, route_path: str):
        payload = await request.json()
        if route_path == "/v1/chat/completions":
            msgs = payload['messages']
            model = payload['model']
            return {
                "host": request.client.host,
                "model": model,
                "messages": [{msg['role']: msg['content']} for msg in msgs],
            }
        else:
            return {}

    @staticmethod
    def parse_bytes_to_content(bytes_: bytes, route_path: str):
        if route_path == "/v1/chat/completions":
            return parse_chat_completions(bytes_)
        else:
            return {}

    def add_chat(self, chat_info: dict):
        logger.info(str(chat_info))
        self._chat_list.append(chat_info)
        self._cur_chat_file_size += 1
        self._save_chat()

    def _save_chat(self):
        if len(self._chat_list) >= self._save_interval:
            logger.info(f"save chat to {self.chat_file}")
            if self._cur_chat_file_size > self._max_chat_file_size:
                logger.info(f"{self._cur_chat_file_size} is too large, create new file")
                self._file_idx += 1
                self._cur_chat_file_size = 1
            self.dump_chat_list(self._chat_list, self.chat_file, mode='a+', _end='\n')
            self._chat_list = []

    @staticmethod
    def dump_chat_list(
        data: List[Dict], filepath: str, rel=False, mode='w', _sep='\n', _end="\n"
    ):
        str_data = _sep.join([str(i) for i in data]) + _end
        abs_path = relp(filepath, parents=1) if rel else filepath
        with open(abs_path, mode=mode) as f:
            f.write(str_data)

    @staticmethod
    def load_chat_list(filepath: str, rel=False, mode='r', _sep='\n'):
        abs_path = relp(filepath, parents=1) if rel else filepath
        with open(abs_path, mode=mode, encoding='utf-8') as f:
            str_result = f.read()
        result_list = str_result.split(_sep)
        result = [eval(i) for i in result_list if i]
        return result



import logging
import os
import sys
import time
from typing import Dict, List, Union

import orjson
from loguru import logger
from rich import print
from rich.panel import Panel
from rich.table import Table
from sparrow import relp


def print_startup_info(base_url, route_prefix, api_key, forward_key, log_chat):
    try:
        from dotenv import load_dotenv

        load_dotenv('.env')
    except Exception:
        ...
    route_prefix = route_prefix or "/"
    api_key_info = True if len(api_key) else False
    forward_key_info = True if len(forward_key) else False
    table = Table(title="", box=None, width=100)
    table.add_column("base-url", justify="left", style="#df412f")
    table.add_column("route-prefix", justify="center", style="#df412f")
    table.add_column("openai-api-key", justify="center", style="green")
    table.add_column("forward-key", justify="center", style="green")
    table.add_column("Log-chat", justify="center", style="green")
    table.add_column("Log-dir", justify="center", style="#f5bb00")
    table.add_row(
        base_url,
        route_prefix,
        str(api_key_info),
        str(forward_key_info),
        str(log_chat),
        "./Log/*.log",
    )
    print(Panel(table, title="🤗openai-forward is ready to serve!", expand=False))


class InterceptHandler(logging.Handler):
    def emit(self, record):
        # Get corresponding Loguru level if it exists
        try:
            level = logger.level(record.levelname).name
        except ValueError:
            level = record.levelno

        # Find caller from where originated the logged message
        frame, depth = logging.currentframe(), 6
        while frame.f_code.co_filename == logging.__file__:
            frame = frame.f_back
            depth += 1
        logger.opt(depth=depth, exception=record.exc_info).log(
            level, record.getMessage()
        )


def setting_log(save_file=False, log_name="openai_forward", multi_process=True):
    # TODO 修复时区配置
    if os.environ.get("TZ") == "Asia/Shanghai":
        os.environ['TZ'] = "UTC-8"
        if hasattr(time, 'tzset'):
            time.tzset()

    logging.root.handlers = [InterceptHandler()]
    for name in logging.root.manager.loggerDict.keys():
        logging.getLogger(name).handlers = []
        logging.getLogger(name).propagate = True

    config_handlers = [
        {"sink": sys.stdout, "level": "DEBUG"},
    ]
    if save_file:
        config_handlers += [
            {
                "sink": f"./Log/{log_name}.log",
                "enqueue": multi_process,
                "rotation": "100 MB",
                "level": "INFO",
            }
        ]

    logger_config = {"handlers": config_handlers}
    logger.configure(**logger_config)


def yaml_dump(data, filepath, rel_path=False, mode='w'):
    abs_path = relp(filepath, parents=1) if rel_path else filepath
    from yaml import dump

    try:
        from yaml import CDumper as Dumper
    except ImportError:
        from yaml import Dumper
    with open(abs_path, mode=mode, encoding="utf-8") as fw:
        fw.write(dump(data, Dumper=Dumper, allow_unicode=True, indent=4))


def yaml_load(filepath, rel_path=False, mode='r'):
    abs_path = relp(filepath, parents=1) if rel_path else filepath
    from yaml import load

    try:
        from yaml import CLoader as Loader
    except ImportError:
        from yaml import Loader
    with open(abs_path, mode=mode, encoding="utf-8") as stream:
        #     stream = stream.read()
        content = load(stream, Loader=Loader)
    return content


def json_load(filepath: str, rel=False, mode='rb'):
    abs_path = relp(filepath, parents=1) if rel else filepath
    with open(abs_path, mode=mode) as f:
        return orjson.loads(f.read())


def json_dump(
    data: Union[List, Dict], filepath: str, rel=False, indent_2=False, mode='wb'
):
    orjson_option = 0
    if indent_2:
        orjson_option = orjson.OPT_INDENT_2
    abs_path = relp(filepath, parents=1) if rel else filepath
    with open(abs_path, mode=mode) as f:
        f.write(orjson.dumps(data, option=orjson_option))


def str2list(s: str, sep=' '):
    if s:
        return [i.strip() for i in s.split(sep) if i.strip()]
    else:
        return []


def env2list(env_name: str, sep=" "):
    return str2list(os.environ.get(env_name, "").strip(), sep=sep)



import os
from itertools import cycle

import httpx
from fastapi import HTTPException, Request, status
from fastapi.responses import StreamingResponse
from loguru import logger
from starlette.background import BackgroundTask

from .config import env2list, print_startup_info, setting_log
from .content.chat import ChatSaver


class OpenaiBase:
    BASE_URL = os.environ.get("OPENAI_BASE_URL", "https://api.openai.com").strip()
    ROUTE_PREFIX = os.environ.get("ROUTE_PREFIX", "").strip()
    _LOG_CHAT = os.environ.get("LOG_CHAT", "False").strip().lower() == "true"
    _openai_api_key_list = env2list("OPENAI_API_KEY", sep=" ")
    _cycle_api_key = cycle(_openai_api_key_list)
    _FWD_KEYS = env2list("FORWARD_KEY", sep=" ")
    _use_forward_key = _openai_api_key_list != [] and _FWD_KEYS != []
    IP_WHITELIST = env2list("IP_WHITELIST", sep=" ")
    IP_BLACKLIST = env2list("IP_BLACKLIST", sep=" ")

    if ROUTE_PREFIX:
        if ROUTE_PREFIX.endswith('/'):
            ROUTE_PREFIX = ROUTE_PREFIX[:-1]
        if not ROUTE_PREFIX.startswith('/'):
            ROUTE_PREFIX = '/' + ROUTE_PREFIX
    timeout = 30

    print_startup_info(
        BASE_URL, ROUTE_PREFIX, _openai_api_key_list, _FWD_KEYS, _LOG_CHAT
    )
    if _LOG_CHAT:
        setting_log(save_file=False)
        chatsaver = ChatSaver(save_interval=10)

    def validate_request_host(self, ip):
        if self.IP_WHITELIST and ip not in self.IP_WHITELIST:
            raise HTTPException(
                status_code=status.HTTP_403_FORBIDDEN,
                detail=f"Forbidden, ip={ip} not in whitelist!",
            )
        if self.IP_BLACKLIST and ip in self.IP_BLACKLIST:
            raise HTTPException(
                status_code=status.HTTP_403_FORBIDDEN,
                detail=f"Forbidden, ip={ip} in blacklist!",
            )

    @classmethod
    async def aiter_bytes(cls, r: httpx.Response, route_path: str):
        bytes_ = b''
        async for chunk in r.aiter_bytes():
            bytes_ += chunk
            yield chunk
        try:
            target_info = cls.chatsaver.parse_bytes_to_content(bytes_, route_path)
            cls.chatsaver.add_chat({target_info['role']: target_info['content']})
        except Exception as e:
            logger.debug(f"log chat (not) error:\n{e=}")

    @classmethod
    async def _reverse_proxy(cls, request: Request):
        client = httpx.AsyncClient(base_url=cls.BASE_URL, http1=True, http2=False)
        url_path = request.url.path
        url_path = url_path[len(cls.ROUTE_PREFIX) :]
        url = httpx.URL(path=url_path, query=request.url.query.encode('utf-8'))
        headers = dict(request.headers)
        auth = headers.pop("authorization", None)
        if auth and str(auth).startswith("Bearer sk-"):
            tmp_headers = {'Authorization': auth}
        elif cls._openai_api_key_list:
            logger.info(f"Use forward key: {cls._use_forward_key}")
            if cls._use_forward_key:
                fk_prefix = "Bearer fk-"
                logger.info(f"current forward key: {auth}")
                if (
                    auth
                    and str(auth).startswith(fk_prefix)
                    and auth[len("Bearer ") :] in cls._FWD_KEYS
                ):
                    auth = "Bearer " + next(cls._cycle_api_key)
                    tmp_headers = {'Authorization': auth}
                else:
                    tmp_headers = {}
            else:
                auth = "Bearer " + next(cls._cycle_api_key)
                tmp_headers = {'Authorization': auth}
        else:
            tmp_headers = {}

        log_chat_completions = False
        if cls._LOG_CHAT and request.method == 'POST':
            try:
                chat_info = await cls.chatsaver.parse_payload_to_content(
                    request, route_path=url_path
                )
                if chat_info:
                    cls.chatsaver.add_chat(chat_info)
                    log_chat_completions = True
            except Exception as e:
                logger.debug(
                    f"log chat error:\n{request.client.host=} {request.method=}: {e}"
                )

        tmp_headers.update({"Content-Type": "application/json"})
        req = client.build_request(
            request.method,
            url,
            headers=tmp_headers,
            content=request.stream(),
            timeout=cls.timeout,
        )
        try:
            r = await client.send(req, stream=True)
        except (httpx.ConnectError, httpx.ConnectTimeout) as e:
            error_info = (
                f"{type(e)}: {e} | "
                f"Please check if host={request.client.host} can access [{cls.BASE_URL}] successfully?"
            )
            logger.error(error_info)
            raise HTTPException(
                status_code=status.HTTP_504_GATEWAY_TIMEOUT, detail=error_info
            )
        except Exception as e:
            error_info = f"{type(e)}: {e}"
            logger.error(error_info)
            raise HTTPException(
                status_code=status.HTTP_500_INTERNAL_SERVER_ERROR, detail=error_info
            )

        aiter_bytes = (
            cls.aiter_bytes(r, url_path) if log_chat_completions else r.aiter_bytes()
        )
        return StreamingResponse(
            aiter_bytes,
            status_code=r.status_code,
            media_type=r.headers.get("content-type"),
            background=BackgroundTask(r.aclose),
        )
